import dataclasses
import logging
import os
import pathlib
import tomllib
from typing import Dict, Iterable, List, Union

import rgpio
import tomli_w

LOGGER = logging.getLogger(__name__)

# Max number of chips to scan for
MAX_CHIPS = 16

@dataclasses.dataclass(frozen=True)
class Chip:
	handle: int
	label: str
	name: str
	number: int
	number_of_lines: int

	@property
	def id(self) -> str:
		return f"{self.number}-{self.name}"

@dataclasses.dataclass(frozen=True)
class Pin:
	chip: Chip
	flags: int
	line_number: int
	name: str
	user: str

	@property
	def id(self) -> str:
		return f"{self.chip.id}-{self.line_number}"

@dataclasses.dataclass(frozen=True)
class Relay:
	pin: Pin
	name: str
	state: bool

	@property
	def id(self) -> str:
		return f"{self.pin.id}"

def _deserialize(data: List[Dict[str, str]]) -> List[Relay]:
	"Deserialize a list of relays from the config."
	return [Relay(
		name=relay["name"],
		pin=Pin(
			chip=relay["chip"],
			number=int(relay["pinnumber"]),
			name=relay["pinname"],
		),
		state=False) for relay in data]

def _serialize(data: List[Relay]) -> List[Dict[str, str]]:
	"Serialize a list of relays to the config."
	LOGGER.info("Serializing %s", data)
	return [{
		"chip": relay.pin.chip,
		"name": relay.name,
		"pinnumber": str(relay.pin.number),
		"pinname": relay.pin.name,
	} for relay in data]

class Manager:
	"Class for interacting with relays."
	def __init__(self, configpath: pathlib.Path, has_fakes: bool):
		try:
			self.config = _read_config(configpath)
		except FileNotFoundError:
			self.config = {
				"rgpio": {"host": "localhost", "port": 8889},
				"relays": [{
					"chip": -1,
					"pin": "FAKE-P1",
					"name": "Fake",
				}],
			}
			# Immediately write out the default config
			_write_config(configpath, self.config)
		self.configpath = configpath
		self.has_fakes = has_fakes

		# Handles to various chips
		self._chips = []
		# Info on various pins
		self._pins = []
		# Cached info on the relays
		self._relays = []
		# Connection to single-board computer GPIO
		self.sbc = None

	def __iter__(self):
		"Provide an iterator for the relays."
		return iter(self.relays)

	@property
	def chips(self) -> List[Chip]:
		if not self._chips:
			self._load_chips()
		return self._chips

	def connect(self) -> None:
		self.sbc = rgpio.sbc(
			host=self.config["rgpio"]["host"],
			port=self.config["rgpio"]["port"],
		)
		if not self.sbc.connected:
			LOGGER.warning("Failed to connect sbc")
			self.sbc = None

	def get_chip_by_number(self, number: int) -> Chip:
		for chip in self.chips:
			if chip.number == number:
				return chip
		raise Exception(f"Can't find chip {number}")

	def get_pin_by_number(self, chip: Chip, line_number: int) -> Pin:
		for pin in self.pins:
			if pin.chip == chip and pin.line_number == line_number:
				return pin
		raise Exception(f"Can't find pin {chip.name}-{line_number}")

	def get_relay_by_pin_id(self, pin_id: str) -> Relay:
		for relay in self.relays:
			if relay.pin.id == pin_id:
				return relay
		return None

	@property
	def pins(self) -> List[Pin]:
		if not self._pins:
			self._load_pins()
		return self._pins

	@property
	def relays(self) -> Iterable[Relay]:
		if not self._relays:
			for relay in self.config["relays"]:
				LOGGER.debug("Relay: %s", relay)
				chip = self.get_chip_by_number(relay["chip"]["number"])
				pin = self.get_pin_by_number(chip, relay["pin"]["line_number"])
				self._relays.append(Relay(
					name=relay["name"],
					pin=pin,
					state=False,
				))
		return self._relays

	def relay_add(self, chip_number: int, chip_name: str, line_number: int, name: str) -> None:
		LOGGER.info("Creating relay %s with chip %d %s on line %d", name, chip_number, chip_name, line_number)
		chip = self.get_chip_by_number(chip_number)
		pin = self.get_pin_by_number(chip, line_number)
		self.config["relays"].append({
			"chip": {
				"name": chip.name,
				"number": chip.number,
			},
			"pin": {
				"name": pin.name,
				"line_number": pin.line_number,
			},
			"name": name,
		})
		_write_config(self.configpath, self.config)

	def relay_set(self, relay: Relay, state: bool) -> None:
		if not relay.pin.chip.startswith("fake"):
			pass

	def shutdown(self) -> None:
		_write_config(self.configpath, self.config)

	# Talk to the remote pins daemon and get information about the chips
	def _load_chips(self) -> None:
		for i in range(MAX_CHIPS):
			try:
				handle = self.sbc.gpiochip_open(i)
			except rgpio.error:
				continue
			if handle < 0:
				continue
			okay, number_of_lines, name, label = self.sbc.gpio_get_chip_info(handle)
			LOGGER.info("Chip info: %s %s %s %s", okay, number_of_lines, name, label)
			if okay != 0:
				LOGGER.warn("Chip %s not okay.", name)
				continue
			self._chips.append(Chip(
				handle=handle,
				label=label,
				name=name,
				number=i,
				number_of_lines=number_of_lines,
			))

	# Talk to the remote pins daemon and get information about the pins
	def _load_pins(self) -> None:
		LOGGER.info("Loading pins")
		for chip in self.chips:
			for line in range(chip.number_of_lines):
				okay, offset, flags, name, user = self.sbc.gpio_get_line_info(chip.handle, line)
				LOGGER.info("Got line info: %s %s %s %s %s", okay, offset, flags, name, user)
				assert offset == line
				if okay != 0:
					LOGGER.warn("Line %s is not okay", name)
					continue
				if not name:
					LOGGER.warn("Ignoring line %d because it has no name.", line)
					continue
				self._pins.append(Pin(
					chip=chip,
					flags=flags,
					line_number=line,
					name=name,
					user=user,
				))


def _read_config(configpath: pathlib.Path) -> None:
	with open(configpath, "rb") as f:
		content = tomllib.load(f)
		LOGGER.info("Read config from %s", configpath)
		return content

def _write_config(configpath: pathlib.Path, config) -> None:
	with open(configpath, "wb") as f:
		tomli_w.dump(config, f)
	LOGGER.info("Wrote config file %s", configpath)

